import sys
from dynamicmethod import dynamicmethod
from pybk8500.__meta__ import version as __version__


__all__ = ['ChecksumError', 'MessageTypeError', 'Parser']


class ChecksumError(Exception):
    pass


class MessageTypeError(Exception):
    pass


class Parser(object):
    NAME = 'pybk8500'
    VERSION = __version__
    MAX_BUFFER_SIZE = 2**16

    START_BYTE = 0xAA
    COMMAND_INDEX = 2
    MSG_LENGTH = 26

    msg_types_lookup = {}

    def __init__(self, name=None):
        # Copy the message types lookup into the instance that can be modified without the class being modified
        self.msg_types_lookup = self.__class__.msg_types_lookup.copy()

        if name is None:
            name = self.NAME

        self._buffer = bytearray()
        self.name = name
        self.error = Parser.error

    @classmethod
    def startindex(cls, data):
        """Return the start index if found. -1 if not found.

        Args:
            data (bytes): Data that you want to find the start of a message for

        Returns:
            index (int): Index of the start position. -1 if not found
        """
        return data.find(cls.START_BYTE)

    @classmethod
    def stopindex(cls, data):
        """Return the stop index if found. -1 if not found.

        Args:
            data (bytes): Data that you want to find the end of a message for. data[0] should be the start index

        Returns:
            index (int): Index of the stop position. -1 if not found
        """
        if len(data) >= cls.MSG_LENGTH:
            return cls.MSG_LENGTH
        else:
            return -1

    @dynamicmethod
    def gen_msg(self, msg_id, raw=None, address=None, **kwargs):
        """Generate a message and return the message bytes.

        It is suggested that this makes a message and sends that message into Parser.to_bytes() to reuse code.

        Returns:
            msg (Message/bytes): Message object
        """
        msg_type = self.lookup(msg_id)
        return msg_type(raw=raw, address=address, **kwargs)

    @staticmethod
    def calc_checksum(byts):
        """Calculate the checksum for the message."""
        byts = byts[:25]  # Do not use the checksum to calculate the checksum
        value = sum(byts)
        return value & 0xFF

    def process_msg(self, byts):
        """Callback to take the parsed message data and create a message object.

        If a message object is successfully created `message_parsed` is called else `error` is called.

        Args:
             byts (bytes): Bytes to convert to a message object and call message_parsed with.
        """
        msg_id = byts[self.COMMAND_INDEX]
        msg_type = self.lookup(msg_id)
        if msg_type is None:
            raise MessageTypeError('Invalid command type {}'.format(msg_id))

        # Create the message from the raw data
        msg = msg_type(raw=byts)

        # Calculate the Checkums
        calc_check = self.calc_checksum(byts)
        if calc_check != msg.checksum:
            raise ChecksumError('The calculated checksum does not match ({} != {})!'.format(calc_check, msg.checksum))

        return msg

    @staticmethod
    def message_parsed(msg):
        """Default "parse" callback function if one is not given."""
        print(msg)

    @staticmethod
    def error(error):
        """Callback to indicate that an error happened.

        Args:
            error (Exception): Optional error object if applicable (C parsers do not create error objects).
        """
        print('{}: {}'.format(type(error).__name__, error), file=sys.stderr)

    def parse(self,  byts, callback=None):
        """Parse the messages in the given bytes use the given callback to use the messages.

        Args:
            byts (bytes): Data to handle
            callback (function)[None]: Function that takes in a parsed message when the message is parsed.

        Returns:
            buffer (bytes): Unused bytes
        """
        if callback is None:
            callback = self.message_parsed

        for msg, remain in self.parse_iter(byts):
            callback(msg)

    def parse_iter(self, byts):
        """Parse the messages in the given bytes.

        Handle errors by overriding the Parser().error callback function.

        Args:
            byts (bytes): Message bytes

        Yields:
            message (Message): The message that was parsed.
            remain (bytes): Remaining bytes
        """
        self._buffer.extend(byts)
        remain = self._buffer

        while True:
            msg, error, remain = self.parse_msg(remain)
            if msg is not None:
                yield msg, remain
            elif error is None:
                break

        self._buffer = bytearray(remain)
        self.check_buffer_size()

    def parse_msg(self, byts):
        """Parse a single message and return the remaining bytes.

        Args:
            byts (bytes): Message bytes

        Returns:
            message (Message)[None]: The message that was parsed.
            error (Exception)[None]: Error that occurred.
            remaining (bytes)[b'']: The remaining bytes
        """
        msg = None
        error = None

        # Find the start index
        startindex = self.startindex(byts)
        if startindex == -1:
            # Trim the buffer, but keep the last few bytes to check for an incomplete start byte
            return msg, error, byts[-int(self.MSG_LENGTH - 1):]

        # Move the buffer
        byts = byts[startindex:]

        # Look for a stop index to see if we have a full message
        stopindex = self.stopindex(byts)
        if stopindex == -1:
            return msg, error, byts

        # Try parsing the message bytes
        try:
            msg = self.process_msg(bytes(byts[0: stopindex]))
            byts = byts[stopindex:]
        except (ChecksumError, MessageTypeError, Exception) as err:
            # Move remaining bytes only by 1 byte. Could be misaligned start byte
            msg = None
            error = err
            byts = byts[1:]
            self.error(error)

        return msg, error, byts

    def check_buffer_size(self):
        """Check the buffer size and limit the amount of data that we hold."""
        if len(self._buffer) > self.MAX_BUFFER_SIZE:
            try:
                raise OverflowError("Overflow Error: Too much data has been given without a valid message!")
            except OverflowError as error:
                self.error(error)
            self._buffer = self._buffer[-int(self.MSG_LENGTH - 1):]

    def flush(self):
        """Flush the buffer."""
        self._buffer = bytearray()

    # ========== Helper Methods ==========
    @dynamicmethod
    def lookup(self, msg_id):
        """Return the message class type for the given message identifier.

        Args:
            msg_id (str/int/bytes/MessageType): Some form of message identifier.

        Returns:
            msg_type (MessageType/callable): MessageType class or function that creates a message with a decode method.
        """
        return self.msg_types_lookup.get(msg_id, None)

    @dynamicmethod
    def add_lookup(self, msg_type, msg_id=None, msg_name=None):
        """Add a custom external message type (that the Packet can parse) to the lookup function.

        Args:
            msg_type (MessageType/class): Message type class that you want to add
            msg_id (str/bytes/int)[None]: Message ID. If this is not given it will use msg_type.ID
            msg_name (str)[None]: Message type display name. If this is not given it will use msg_type.NAME
        """
        if msg_id is None:
            try:
                msg_id = msg_type.ID
            except AttributeError:
                pass
        if msg_name is None:
            try:
                msg_name = msg_type.NAME
            except AttributeError:
                pass

        self.msg_types_lookup[msg_type] = msg_type
        if msg_id:
            self.msg_types_lookup[msg_id] = msg_type
        if msg_name:
            self.msg_types_lookup[msg_name] = msg_type
        return msg_type

    @dynamicmethod
    def remove_lookup(self, msg_type=None, msg_id=None, msg_name=None):
        """Remove a message type from the lookup.

        Args:
            msg_type (MessageType/class)[None]: Message type class that you want to add
            msg_id (str/bytes/int)[None]: Message ID. If this is not given it will use msg_type.ID
            msg_name (str)[None]: Message type display name. If this is not given it will use msg_type.NAME
        """
        if msg_id is None:
            try:
                msg_id = msg_type.ID
            except AttributeError:
                pass
        if msg_name is None:
            try:
                msg_name = msg_type.NAME
            except AttributeError:
                pass

        try:
            msg = None
            if msg_name:
                msg = self.msg_types_lookup.pop(msg_name, msg)
            if msg_id:
                msg = self.msg_types_lookup.pop(msg_id, msg)
            if msg_type:
                msg = self.msg_types_lookup.pop(msg_type, msg)
            return msg
        except KeyError:
            pass

    def __getstate__(self):
        """Return the objects variables for serialization/pickling, so this object can be recreated in a
        separate process.
        """
        d = {'name': self.name}
        if self.msg_types_lookup != self.__class__.msg_types_lookup:
            d['msg_types_lookup'] = self.msg_types_lookup
        if self.error != self.__class__.error:
            d['error'] = self.error
        return d

    def __setstate__(self, state):
        """After deserialization/pickling in a separate process reset this object's values from the state."""
        self.msg_types_lookup = self.__class__.msg_types_lookup.copy()
        self._buffer = bytearray()
        self.name = self.NAME
        self.error = Parser.error

        for k, v in state.items():
            setattr(self, k, v)



import re
import datetime


__all__ = ['parse_number', 'TIME_FORMATS', 'NUMBER_UNIT_REGEX', 'UNIT_CONVERT']


TIME_FORMATS = [
    '%I:%M:%S %p',     # '02:24:55 PM'
    '%I:%M:%S.%f %p',  # '02:24:55.000200 PM'
    '%I:%M %p',        # '02:24 PM'
    '%H:%M:%S',        # '14:24:55'
    '%H:%M:%S.%f',     # '14:24:55.000200'
    '%H:%M',           # '14:24'
    ]


NUMBER_UNIT_REGEX = re.compile(r'^(?P<num>\d*\.?\d*) ?(?P<unit>\w*)')
UNIT_CONVERT = {
    'T': 10**12, 'Terra': 10**12,
    'G': 10**9, 'Giga': 10**9,
    'M': 10**6, 'Mega': 10**6,
    'k': 10**3, 'kilo': 10**3,
    'c': 10**-2, 'centi': 10**-2,
    'm': 10**-3, 'milli': 10**-3,
    'µ': 10**-6, 'micro': 10**-6,
    'n': 10**-9, 'nano': 10**-9,
    'p': 10**-12, 'pico': 10**-12,
    'min': 60, 'minute': 60, 'minutes': 60,
    'h': 60*60, 'hour': 60*60, 'hours': 60*60,
    }


def parse_number(value):
    """Try parsing the value into a number.

    If units are given convert to the base unit. If HH:MM:SS then return the number of seconds.
    """
    if ":" in str(value):
        for fmt in TIME_FORMATS:
            try:
                dt = datetime.datetime.strptime(value, fmt)
                return (dt - datetime.datetime(1900, 1, 1)).total_seconds()
            except (ValueError, TypeError, Exception):
                pass

    # Check regex
    d = NUMBER_UNIT_REGEX.match(value).groupdict()
    num = d['num']
    unit = d['unit']
    if not num:
        return value

    # Convert to number
    try:
        num = int(num)
    except (ValueError, TypeError, Exception):
        try:
            num = float(num)
        except (ValueError, TypeError, Exception):
            return num  # Cannot get a real number

    # Find unit converter
    for name, scale in UNIT_CONVERT.items():
        if unit.startswith(name):
            try:
                return num * scale
            except (ValueError, TypeError, Exception):
                pass

    return num



"""
Manage communications between the PC and the device.

Requirements:
    * pyserial
    * continuous_threading

"""
import sys
import time
import serial
import functools
import contextlib
import continuous_threading
from pybk8500.parser import Parser


__all__ = ['CommunicationManager', 'send_msg', 'main']


class CommunicationManager(object):
    Parser = Parser
    DEFAULT_READ_SIZE = 26

    def __init__(self, connection=None, parser=None, com=None, baudrate=None, **kwargs):
        super().__init__()

        if parser is None:
            parser = self.Parser()
        if connection is None:
            connection = serial.Serial()
            if com is not None:
                connection.port = com
            if baudrate is not None:
                connection.baudrate = baudrate
            # connection.rts = True  # Documentation states needed. Did not work
            # connection.dtr = True  # Documentation states needed. Did not work

        self._parser = None
        self._process = None
        self._in_enter = False
        self._enter_started = False
        self._enter_connected = False
        self.read_delay = 0.0001
        self.wait_delay = 0.01
        self.read_size = None  # Use in_waiting if None

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.ack_list = []
        self.response_types = []
        self.connection = connection
        self.set_parser(parser)

    def get_parser(self):
        """Return the parser."""
        return self._parser

    def set_parser(self, parser):
        """Set the parser.

        Args:
            parser (object/None/Parser)[None]: Parser object to parse incoming messages.
        """
        self._parser = parser
        if self._parser is not None:
            self._parser.message_parsed = self.message_parsed
            self._parser.error = self.error

    parser = property(get_parser, set_parser)

    def save_ack(self, msg):
        """Save the response messages in the available response_types."""
        if len(self.response_types) == 0 or any(isinstance(msg, rtype) for rtype in self.response_types):
            self.ack_list.append(msg)

    message_parsed = save_ack

    @contextlib.contextmanager
    def change_message_parsed(self, callback):
        """Change the message parsed function while in this with block."""
        old = self.message_parsed
        self.message_parsed = callback

        yield

        self.message_parsed = old

    @staticmethod
    def error(error):
        """Callback to indicate that an error happened.

        Args:
            error (Exception): Optional error object if applicable (C parsers do not create error objects).
        """
        print('{}: {}'.format(type(error).__name__, error), file=sys.stderr)

    @contextlib.contextmanager
    def change_connection(self):
        """Change the connection properties safely."""
        is_connected = self.is_connected()
        if is_connected:
            self.disconnect()

        yield

        if is_connected:
            self.connect()

    def get_baudrate(self):
        """Return the baudrate."""
        return self.connection.baudrate

    def set_baudrate(self, value, *args, **kwargs):
        """Set the baudrate."""
        with self.change_connection():
            self.connection.baudrate = value

    def get_com(self):
        """Return the serial com port."""
        return self.connection.port

    def set_com(self, value, *args, **kwargs):
        """Set the serial com port and try to connect."""
        with self.change_connection():
            self.connection.port = value

        if not self.is_connected():
            try:
                self.connect()
                if self._in_enter:
                    self._enter_connected = True
            except Exception as err:
                print('Warning: Could not connect! {}'.format(err), file=sys.stderr)

    get_port = get_com
    set_port = set_com

    def get_rts(self):
        """Return if the RTS Hardware Flow Control is set."""
        try:
            return self.connection.rts
        except (AttributeError, Exception):
            return False

    def set_rts(self, value, *args, **kwargs):
        """Set the RTS Hardware Flow Control."""
        with self.change_connection():
            self.connection.rts = bool(value)

    def get_dtr(self):
        """Return if the DTR Hardware Flow Control is set."""
        try:
            return self.connection.dtr
        except (AttributeError, Exception):
            return False

    def set_dtr(self, value, *args, **kwargs):
        """Set the DTR Hardware Flow Control."""
        with self.change_connection():
            self.connection.dtr = bool(value)

    def is_connected(self):
        """Return if the connection/serial port is connected."""
        try:
            if isinstance(self.connection, serial.Serial):
                return self.connection.is_open
        except (AttributeError, Exception):
            pass
        return False

    def connect(self, com=None, baudrate=None, **kwargs):
        """Connect the connection/serial port."""
        if com is not None or baudrate is not None:
            self.disconnect()
        if com is not None:
            self.connection.port = com
        if baudrate is not None:
            self.connection.baudrate = baudrate

        if not self.is_connected():
            self.flush()
            if isinstance(self.connection, serial.Serial):
                self.connection.open()

    def disconnect(self, *args, **kwargs):
        """Disconnect the connection/serial port."""
        if isinstance(self.connection, serial.Serial):
            self.connection.close()

    def flush(self):
        """Clear the message buffer and input buffer."""
        self.ack_list.clear()
        try:
            self.connection.flush()
        except (AttributeError, Exception):
            pass
        try:
            self.connection.reset_input_buffer()
        except (AttributeError, Exception):
            pass
        try:
            self.connection.reset_output_buffer()
        except (AttributeError, Exception):
            pass

    def read(self):
        """Read data from the connection."""
        if isinstance(self.connection, serial.Serial):
            read_size = self.read_size
            if read_size is None or read_size <= 0:
                read_size = self.connection.in_waiting
            return self.connection.read(read_size)
        else:
            return b''

    def write(self, byts):
        """Write the bytes (or message) data to the connection."""
        return self.connection.write(bytes(byts))

    def read_and_parse(self):
        """Read data from the connection and parse it."""
        try:
            if self.is_connected():
                byts = self.read()
                if byts:
                    self.parser.parse(byts, self.message_parsed)
                time.sleep(self.read_delay)
            else:
                time.sleep(0.1)
        except (ConnectionAbortedError, Exception) as err:
            print(str(err), file=sys.stderr)

    @contextlib.contextmanager
    def listen_for_messages(self, *msg_types):
        """Context manager to listen for certain message types."""
        # Ensure connected and running
        is_connected = self.is_connected()
        is_running = self.is_running()
        if not is_connected:
            self.connect()
        if not is_running:
            self.start()

        # Start listening for responses
        for msg_type in msg_types:
            if msg_type is not None:
                self.response_types.append(msg_type)

        try:
            # Yield with block
            yield

        finally:
            # Remove message types
            for msg_type in msg_types:
                try:
                    self.response_types.remove(msg_type)
                except (KeyError, IndexError, Exception):
                    pass

            # If connected and/or started then stop and/or disconnect
            if not is_running:
                self.stop()
            if not is_connected:
                self.disconnect()

    @classmethod
    def listener(cls, *msg_types, attr=None, func=None):
        """Decorator to have a function run with listen_for_messages"""
        if func is None:
            def decorator(f):
                return cls.listener(*msg_types, func=f)
            return decorator

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            msg_mngr = self
            if attr is not None:
                msg_mngr = getattr(self, msg_mngr)
            with msg_mngr.listen_for_messages(*msg_types):
                return func(self, *args, **kwargs)

        return wrapper

    def wait_for_response(self, timeout, msg_type=None):
        """Wait for a response message and return True if a message was received.

        Args:
            timeout (float/int): Number of seconds to wait for a message.
            msg_type (Message/object)[None]: Message type class to wait for.

        Returns:
            success (bool): True if a message was received within the timeout.
        """
        start = time.time()
        while (time.time() - start) < timeout:
            if (msg_type is None and len(self.ack_list) > 0) or any(isinstance(msg, msg_type) for msg in self.ack_list):
                return True
            time.sleep(self.wait_delay)
        return False

    def send_wait(self, msg, timeout=0, msg_type=None, attempts=3, print_msg=True, print_recv=True):
        """Send a message and wait for a response.

        Args:
            msg (Message): Message to convert to bytes and send.
            timeout (float/int): Number of seconds to wait for a message on each attempt.
            msg_type (Message/object)[None]: Message type class to wait for.
            attempts (int)[3]: Number of attempts to send the message and wait for the response.
            print_msg (bool)[True]: If True print out that you are sending the message.
            print_recv (bool)[True]: If True print all received messages.

        Returns:
            ack_list (list): List of received messages.
        """
        with self.listen_for_messages(msg_type):
            trials = 0
            success = False
            pout = 'Sending {} ...'.format(msg)
            while (trials < attempts) and not success:
                if print_msg:
                    print(pout)
                self.write(bytes(msg))
                success = self.wait_for_response(timeout, msg_type=msg_type)
                pout = 'Retry sending {} ...'.format(msg)
                trials += 1

            if not success and timeout > 0:
                raise TimeoutError('Attempts sending {} failed!'.format(msg))

        # Clear and return messages
        msgs = [self.ack_list.pop(i) for i in reversed(range(len(self.ack_list)))
                if msg_type is None or isinstance(self.ack_list[i], msg_type)]
        msgs = reversed(msgs)

        if print_recv:
            for msg in msgs:
                print('Received:', msg)

        return msgs

    send_wait_for_response = send_wait

    def is_running(self):
        """Return if the reading thread is running."""
        return self._process is not None and self._process.is_running()

    def start(self):
        """Start reading and parsing the connection."""
        if self._process is None:
            self._process = continuous_threading.PausableThread(target=self.read_and_parse)

        self.flush()
        self._process.start()

        # Wait for the thread to start reading
        time.sleep(0.01)

        return self

    def stop(self):
        """Stop reading and parsing the connection."""
        try:
            self._process.stop()
        except (AttributeError, Exception):
            pass
        return self

    def close(self):
        """Close the process."""
        self.disconnect()
        try:
            self._process.close()
        except (AttributeError, Exception):
            pass
        self._process = None
        return self

    def __enter__(self):
        """Enter the 'with' context manager."""
        self._in_enter = True
        if not self.is_connected():
            try:
                self.connect()
                self._enter_connected = True
            except Exception as err:
                print('Warning: Could not connect! {}'.format(err), file=sys.stderr)
        if not self.is_running():
            self.start()
            self._enter_started = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the 'with' context manager."""
        self._in_enter = False
        if self._enter_started:
            self._enter_started = False
            self.stop()
        if self._enter_connected:
            self._enter_connected = False
            self.disconnect()
        return exc_type is None


def send_msg(com, baudrate, cmd_id, timeout=1, attempts=1, **kwargs):
    """Send a command to the device.

    Args:
        com (str): Com port to connect to
        baudrate (int): Baudrate to connect with.
        cmd_id (int/str/Message): Command identifier to send.
        timeout (float/int)[1]: Timeout to wait for the response.
        attempts (int)[1]: Number of times to send the message expecting a response.
        **kwargs (dict): Dictionary of Command keyword arguments (variable names with values).
    """
    cmd_type = Parser.lookup(cmd_id)
    if cmd_type is None:
        raise ValueError('Invalid cmd_id given! No matching command for {}'.format(cmd_id))

    cmd = cmd_type(**kwargs)

    with CommunicationManager(com=com, baudrate=baudrate) as ser:
        try:
            msgs = ser.send_wait(cmd, timeout=timeout, msg_type=cmd.RESPONSE_TYPE, attempts=attempts)
        except TimeoutError:
            # Timeout error with no response for the expected type.
            msgs = [ser.ack_list.pop(0) for _ in range(len(ser.ack_list))]

        for msg in msgs:
            print('Received {}:'.format(msg.NAME))
            for field, value in msg.fields().items():
                print('\t{} = {}'.format(field, value))
            print()


main = send_msg


def cli_to_kwargs(cli_args):
    """Convert command line arguments to a dictionary.

    Args:
        cli_args (list): List of command line arguments ["--address", "1", "--value", 2]
    """
    return {get_name(cli_args[i]): get_value(cli_args[i+1]) for i in range(0, len(cli_args), 2)}


def get_name(name):
    """Get a command line argument name by removing all '-'."""
    return str(name).replace('-', '')


def get_value(value):
    """Convert the given string value to a proper python object.

    ast.literal_eval may work better.
    """
    try:
        if str(value).startswith('0x'):
            return int(value, 16)  # Was given hex. This allows "0x12" = 18
        else:
            return int(value)
    except (ValueError, TypeError, Exception):
        try:
            return float(value)
        except (ValueError, TypeError, Exception):
            return value


if __name__ == '__main__':
    import argparse

    P = argparse.ArgumentParser(description='Send a command to the device.')
    P.add_argument('com', type=str, help='Com port to connect to.')
    P.add_argument('baudrate', type=int, help='Baudrate to connect with.')
    P.add_argument('cmd_id', type=str,
                   help='Command ID as the string NAME or integer ID '
                        '(Example: "Command Status" String, 0x12 Hex, or 18 Dec).')
    P.add_argument('--timeout', type=float, default=1, help='Timeout to wait for the response.')
    P.add_argument('--attempts', type=int, default=1, help='Number of times to send the message expecting a response.')
    ARGS, REMAINDER = P.parse_known_args()

    main(ARGS.com, ARGS.baudrate, get_value(ARGS.cmd_id), timeout=ARGS.timeout, attempts=ARGS.attempts,
         **cli_to_kwargs(REMAINDER))
